#define _USE_32BIT_TIME_T
#include <windows.h>
#include <shlwapi.h>
#define TVTEST_PLUGIN_CLASS_IMPLEMENT
#include "TVTestPlugin.h"
#include "IDTVPlugin.h"


// ΉSDK̃o[W
//#define SDK_VERSION 1	// Version 1.3
//#define SDK_VERSION 2	// Version 1.4
//#define SDK_VERSION 3	// Version 1.5
#ifndef SDK_VERSION
#define SDK_VERSION 3
#endif


#pragma comment(lib,"shlwapi.lib")


#define DUMMY_WINDOW_CLASS TEXT("TVBridge Dummy Window %s")


#ifdef _DEBUG
void DebugTrace(LPCTSTR pszFormat, ...)
{
	va_list Args;
	TCHAR szOutput[1024];

	va_start(Args,pszFormat);
	wvsprintf(szOutput,pszFormat,Args);
	va_end(Args);
	OutputDebugString(szOutput);
}
#undef TRACE
#define TRACE DebugTrace
#else
#define TRACE __noop
#endif


class CCriticalLock
{
public:
	CCriticalLock();
	virtual ~CCriticalLock();

	void Lock(void);
	void Unlock(void);
	bool TryLock(DWORD TimeOut=0);
private:
	CRITICAL_SECTION m_CriticalSection;
};

CCriticalLock::CCriticalLock()
{
	::InitializeCriticalSection(&m_CriticalSection);
}

CCriticalLock::~CCriticalLock()
{
	::DeleteCriticalSection(&m_CriticalSection);
}

void CCriticalLock::Lock(void)
{
	::EnterCriticalSection(&m_CriticalSection);
}

void CCriticalLock::Unlock(void)
{
	::LeaveCriticalSection(&m_CriticalSection);
}

bool CCriticalLock::TryLock(DWORD TimeOut)
{
	bool bLocked=false;

	if (TimeOut==0) {
		if (::TryEnterCriticalSection(&m_CriticalSection))
			bLocked=true;
	} else {
		for (DWORD i=TimeOut;i>0;i--) {
			if (::TryEnterCriticalSection(&m_CriticalSection)) {
				bLocked=true;
				break;
			}
			Sleep(1);
		}
	}
	return bLocked;
}


class IDTVApp_
{
public:
	virtual ~IDTVApp_() { }
	virtual BOOL WINAPI put_Channel(int Channel) =0;
	virtual int WINAPI get_Channel(void) =0;
	virtual BOOL WINAPI put_ServiceID(int ProgramNumber) =0;
	virtual int WINAPI get_ServiceID() =0;
	virtual BOOL WINAPI GetProgramList(CDTVProgram *Programs, int *Count) =0;
	virtual void WINAPI TerminateApp(void) =0;
#if SDK_VERSION>=2
	virtual BOOL WINAPI GetStatus(CDTVStatus *Status) =0;
	virtual BOOL WINAPI ParseISO2022Text(BYTE *SrcBuf, int SrcLen, WCHAR *DestBuf, int DestLen) =0;
#if SDK_VERSION>=3
	virtual BOOL WINAPI put_ChannelAndServiceID(int Channel, int Service) =0;
#endif
#endif
};


struct ChannelInfo {
	WORD Frequency;
	WORD TSID;
};

static const ChannelInfo BSChannelList[] = {
	{11727,	0x4010},	// BS1/TS0 BS
	{11727,	0x4011},	// BS1/TS1 BS-i
	{11766,	0x4030},	// BS3/TS0 WOWOW
	{11766,	0x4031},	// BS3/TS1 BSWp
	{11880,	0x4090},	// BS9/TS0 BS11
	{11880,	0x4091},	// BS9/TS1 Star Channel HV
	{11880,	0x4092},	// BS9/TS2 TwellV
	{11958,	0x40D0},	// BS13/TS0 BSe
	{11958,	0x40D1},	// BS13/TS1 BStW
	{11996,	0x40F1},	// BS15/TS1 NHK BS1/2
	{11996,	0x40F2},	// BS15/TS2 NHK BS-hi
};

static const ChannelInfo CSChannelList[] = {
	{12291,	0x6020},	// ND2 110CS #1
	{12331,	0x7040},	// ND4 110CS #2
	{12371,	0x7060},	// ND6 110CS #3
	{12411,	0x6080},	// ND8 110CS #4
	{12451,	0x60A0},	// ND10 110CS #5
	{12491,	0x70C0},	// ND12 110CS #6
	{12531,	0x70E0},	// ND14 110CS #7
	{12571,	0x7100},	// ND16 110CS #8
	{12611,	0x7120},	// ND18 110CS #9
	{12651,	0x7140},	// ND20 110CS #10
	{12691,	0x7160},	// ND22 110CS #11
	{12731,	0x7180},	// ND24 110CS #12
};


class CDTVApp : public IDTVApp_
{
	TVTest::CTVTestApp *m_pApp;
	DWORD m_FriioID;
	int m_TerrestrialIndex;
	int m_BSIndex;
	int m_CSIndex;
public:
	CDTVApp(TVTest::CTVTestApp *pApp,DWORD FriioID);
	void SetSpaceIndex(int Terrestrial,int BS,int CS);
	virtual BOOL WINAPI put_Channel(int Channel);
	virtual int WINAPI get_Channel(void);
	virtual BOOL WINAPI put_ServiceID(int ProgramNumber);
	virtual int WINAPI get_ServiceID();
	virtual BOOL WINAPI GetProgramList(CDTVProgram *Programs,int *Count);
	virtual void WINAPI TerminateApp(void);
#if SDK_VERSION>=2
	virtual BOOL WINAPI GetStatus(CDTVStatus *Status);
	virtual BOOL WINAPI ParseISO2022Text(BYTE *SrcBuf,int SrcLen,WCHAR *DestBuf,int DestLen);
#if SDK_VERSION>=3
	virtual BOOL WINAPI put_ChannelAndServiceID(int Channel,int Service);
#endif
#endif
};


CDTVApp::CDTVApp(TVTest::CTVTestApp *pApp,DWORD FriioID)
	: m_pApp(pApp)
	, m_FriioID(FriioID)
	, m_TerrestrialIndex(0)
	, m_BSIndex(1)
	, m_CSIndex(2)
{
}


void CDTVApp::SetSpaceIndex(int Terrestrial,int BS,int CS)
{
	m_TerrestrialIndex=Terrestrial;
	m_BSIndex=BS;
	m_CSIndex=CS;
}


BOOL WINAPI CDTVApp::put_Channel(int Channel)
{
	TVTest::ChannelInfo ChannelInfo;
	int i,j;

	TRACE(TEXT("TVBridge : CDTVApp::put_Channel(%d)\n"),Channel);
	if (m_FriioID==FRIIO_TERRESTRIAL) {
		for (i=0;m_pApp->GetChannelInfo(m_TerrestrialIndex,i,&ChannelInfo);i++) {
			if (ChannelInfo.PhysicalChannel==Channel)
				return m_pApp->SetChannel(m_TerrestrialIndex,i);
		}
		for (i=0;m_pApp->GetChannelInfo(m_TerrestrialIndex,i,&ChannelInfo);i++) {
			if (ChannelInfo.Channel==Channel-13)
				return m_pApp->SetChannel(m_TerrestrialIndex,i);
		}
	} else if (m_FriioID==FRIIO_SATELLITE) {
		WORD Frequency=Channel>>16,TSID=(Channel&0xFFFF);

		for (i=0;i<sizeof(BSChannelList)/sizeof(ChannelInfo);i++) {
			if (BSChannelList[i].Frequency==Frequency
					&& (TSID==0 || BSChannelList[i].TSID==TSID)) {
				for (j=0;m_pApp->GetChannelInfo(m_BSIndex,j,&ChannelInfo);j++) {
					if (ChannelInfo.Channel==i)
						return m_pApp->SetChannel(m_BSIndex,j);
				}
				return FALSE;
			}
		}
		for (i=0;i<sizeof(CSChannelList)/sizeof(ChannelInfo);i++) {
			if (CSChannelList[i].Frequency==Frequency
					&& (TSID==0 || CSChannelList[i].TSID==TSID)) {
				for (j=0;m_pApp->GetChannelInfo(m_CSIndex,j,&ChannelInfo);j++) {
					if (ChannelInfo.Channel==i)
						return m_pApp->SetChannel(m_CSIndex,j);
				}
				return FALSE;
			}
		}
	}
	return FALSE;
}


int WINAPI CDTVApp::get_Channel(void)
{
	TVTest::ChannelInfo ChannelInfo;

	TRACE(TEXT("TVBridge : CDTVApp::get_Channel()\n"));
	if (!m_pApp->GetCurrentChannelInfo(&ChannelInfo))
		return 0;
	if (m_FriioID==FRIIO_TERRESTRIAL) {
		if (ChannelInfo.PhysicalChannel!=0)
			return ChannelInfo.PhysicalChannel;
		return ChannelInfo.Channel+13;
	}
	if (m_FriioID==FRIIO_SATELLITE) {
		if (ChannelInfo.Space==m_BSIndex) {
			if (ChannelInfo.Channel<sizeof(BSChannelList)/sizeof(ChannelInfo))
				return (BSChannelList[ChannelInfo.Channel].Frequency<<16)|
						BSChannelList[ChannelInfo.Channel].TSID;
		} else if (ChannelInfo.Space==m_CSIndex) {
			if (ChannelInfo.Channel<sizeof(CSChannelList)/sizeof(ChannelInfo))
				return (CSChannelList[ChannelInfo.Channel].Frequency<<16)|
						CSChannelList[ChannelInfo.Channel].TSID;
		}
	}
	return 0;
}


BOOL WINAPI CDTVApp::put_ServiceID(int ProgramNumber)
{
	TRACE(TEXT("TVBridge : CDTVApp::put_ServiceID(%d)\n"),ProgramNumber);
	return m_pApp->SetService(ProgramNumber,true);
}


int WINAPI CDTVApp::get_ServiceID()
{
	int Service=m_pApp->GetService();
	TVTest::ServiceInfo ServiceInfo;

	TRACE(TEXT("TVBridge : CDTVApp::get_ServiceID()\n"));
	if (Service<0 || !m_pApp->GetServiceInfo(Service,&ServiceInfo))
		return 0;
	return ServiceInfo.ServiceID;
}


BOOL WINAPI CDTVApp::GetProgramList(CDTVProgram *Programs,int *Count)
{
	TVTest::ServiceInfo ServiceInfo;
	TVTest::ProgramInfo ProgramInfo;
	CDTVProgram DTVProgram;

	TRACE(TEXT("TVBridge : CDTVApp::get_ProgramList()\n"));
	::ZeroMemory(&DTVProgram,sizeof(CDTVProgram));
	ProgramInfo.pszEventName=DTVProgram.event_name;
	ProgramInfo.MaxEventName=255;
	ProgramInfo.pszEventText=DTVProgram.event_text;
	ProgramInfo.MaxEventText=255;
	ProgramInfo.pszEventExtText=NULL;
	if (!m_pApp->GetServiceInfo(m_pApp->GetService(),&ServiceInfo)
			|| !m_pApp->GetCurrentProgramInfo(&ProgramInfo)) {
		*Count=0;
		return FALSE;
	}
	if (*Count<1) {
		*Count=1;
		return FALSE;
	}
	::lstrcpyn(DTVProgram.service_name,ServiceInfo.szServiceName,32);
	DTVProgram.program_number=ServiceInfo.ServiceID;
	DTVProgram.event_start_time=
		((ProgramInfo.StartTime.wHour/10)<<20)|
		((ProgramInfo.StartTime.wHour%10)<<16)|
		((ProgramInfo.StartTime.wMinute/10)<<12)|
		((ProgramInfo.StartTime.wMinute%10)<<8)|
		((ProgramInfo.StartTime.wSecond/10)<<4)|
		(ProgramInfo.StartTime.wSecond%10);
	DTVProgram.event_duration=
		((ProgramInfo.Duration/(60*60)/10)<<20)|
		((ProgramInfo.Duration/(60*60)%10)<<16)|
		((ProgramInfo.Duration/60/10)<<12)|
		((ProgramInfo.Duration/60%10)<<8)|
		((ProgramInfo.Duration%60/10)<<4)|
		(ProgramInfo.Duration%60%10);
	DTVProgram.vpid=ServiceInfo.VideoPID;
	DTVProgram.apid[0]=ServiceInfo.AudioPID[0];
	*Count=1;
	return TRUE;
}


void WINAPI CDTVApp::TerminateApp(void)
{
	TRACE(TEXT("TVBridge : CDTVApp::TerminateApp()\n"));
	m_pApp->Close(TVTest::CLOSE_EXIT);
}


#if SDK_VERSION>=2


BOOL WINAPI CDTVApp::GetStatus(CDTVStatus *Status)
{
	TVTest::StatusInfo Info;
	TVTest::RecordStatusInfo RecordInfo;

	TRACE(TEXT("TVBridge : CDTVApp::GetStatus()\n"));
	if (!m_pApp->GetStatus(&Info) || !m_pApp->GetRecordStatus(&RecordInfo))
		return FALSE;
	Status->SmartCardStatus=DTV_CARD_OK;
	Status->CN=Info.SignalLevel;
	Status->BitRate=Info.BitRate;
	Status->CCErrorCount=Info.ErrorPacketCount;
	Status->Recording=RecordInfo.Status!=TVTest::RECORD_STATUS_NOTRECORDING;
	if (RecordInfo.Status!=TVTest::RECORD_STATUS_NOTRECORDING) {
		DWORD RecordTime=RecordInfo.RecordTime+RecordInfo.PauseTime;

		Status->RecordTime=RecordTime/1000;
		if (RecordInfo.StopTimeSpec==TVTest::RECORD_STOP_TIME) {
			SYSTEMTIME st;
			FILETIME ft;
			ULARGE_INTEGER Cur,Stop;

			::GetLocalTime(&st);
			::SystemTimeToFileTime(&st,&ft);
			Cur.LowPart=ft.dwLowDateTime;
			Cur.HighPart=ft.dwHighDateTime;
			Stop.LowPart=RecordInfo.StopTime.Time.dwLowDateTime;
			Stop.HighPart=RecordInfo.StopTime.Time.dwHighDateTime;
			if (Cur.QuadPart<Stop.QuadPart)
				Status->RecordRemainTime=(time_t)((Stop.QuadPart-Cur.QuadPart)/10000000);
			else
				Status->RecordRemainTime=0;
		} else if (RecordInfo.StopTimeSpec==TVTest::RECORD_STOP_DURATION) {
			if (RecordTime<=RecordInfo.StopTime.Duration)
				Status->RecordRemainTime=
					(time_t)((RecordInfo.StopTime.Duration-RecordTime)/1000);
			else
				Status->RecordRemainTime=0;
		} else {
			Status->RecordRemainTime=(0xFFFFFFFFUL-RecordTime)/1000;
		}
	} else {
		Status->RecordTime=0;
		Status->RecordRemainTime=0;
	}
	Status->Preview=m_pApp->GetPreview();
	return TRUE;
}


BOOL WINAPI CDTVApp::ParseISO2022Text(BYTE *SrcBuf,int SrcLen,WCHAR *DestBuf,int DestLen)
{
	TRACE(TEXT("TVBridge : CDTVApp::ParseISO2022Text()\n"));
	return m_pApp->DecodeARIBString(SrcBuf,SrcLen,DestBuf,DestLen);
}


#if SDK_VERSION>=3


BOOL WINAPI CDTVApp::put_ChannelAndServiceID(int Channel,int Service)
{
	TRACE(TEXT("TVBridge : CDTVApp::put_ChannelAndServiceID(%d,%d)\n"),Channel,Service);

	TVTest::ChannelInfo ChannelInfo;
	int i,j;

	if (m_FriioID==FRIIO_TERRESTRIAL) {
		for (i=0;m_pApp->GetChannelInfo(m_TerrestrialIndex,i,&ChannelInfo);i++) {
			if ((ChannelInfo.PhysicalChannel==Channel
					|| ChannelInfo.Channel==Channel-13)
					&& (int)ChannelInfo.ServiceID==Service)
				return m_pApp->SetChannel(m_TerrestrialIndex,i);
		}
	} else if (m_FriioID==FRIIO_SATELLITE) {
		WORD Frequency=Channel>>16,TSID=(Channel&0xFFFF);

		for (i=0;i<sizeof(BSChannelList)/sizeof(ChannelInfo);i++) {
			if (BSChannelList[i].Frequency==Frequency
					&& (TSID==0 || BSChannelList[i].TSID==TSID)) {
				for (j=0;m_pApp->GetChannelInfo(m_BSIndex,j,&ChannelInfo);j++) {
					if (ChannelInfo.Channel==i
							&& (int)ChannelInfo.ServiceID==Service)
						return m_pApp->SetChannel(m_BSIndex,j);
				}
				break;
			}
		}
		for (i=0;i<sizeof(CSChannelList)/sizeof(ChannelInfo);i++) {
			if (CSChannelList[i].Frequency==Frequency
					&& (TSID==0 || CSChannelList[i].TSID==TSID)) {
				for (j=0;m_pApp->GetChannelInfo(m_CSIndex,j,&ChannelInfo);j++) {
					if (ChannelInfo.Channel==i
							&& (int)ChannelInfo.ServiceID==Service)
						return m_pApp->SetChannel(m_CSIndex,j);
				}
				break;
			}
		}
	}
	if (!put_Channel(Channel))
		return FALSE;
	return put_ServiceID(Service);
}


#endif	// SDK_VERSION>=3
#endif	// SDK_VERSION>=2


class IDTVPlugin_
{
protected:
	virtual ~IDTVPlugin_() { }
public:
	virtual BOOL WINAPI SetHostAPI(IDTVApp *Application, DWORD Version) =0;
	virtual BOOL WINAPI Init(HWND AppWindow) =0;
	virtual BOOL WINAPI ProcessTS(BYTE *Buffer, DWORD Length) =0;
	virtual BOOL WINAPI OnStreamChange(int Channel, int ServiceID) =0;
#if SDK_VERSION>=2
	virtual BOOL WINAPI OnFileSave(TCHAR *Buffer, DWORD Length) =0;
#endif
};




class CTVBridge : public TVTest::CTVTestPlugin
{
	TCHAR m_szFriioPluginName[MAX_PATH];
	HMODULE m_hLib;
	CDTVApp *m_pDTVApp;
	IDTVPlugin_ *m_pDTVPlugin;
	DWORD m_SDKVersion;
	DWORD m_FriioID;
	int m_TerrestrialIndex;
	int m_BSIndex;
	int m_CSIndex;
	TCHAR m_szClassName[32+MAX_PATH];
	HWND m_hwndDummy;
	DWORD m_RecordTime;
	enum {
		DEFAULT_BUFFER_LENGTH=20000UL,
		DEFAULT_SEND_BUFFER_LENGTH=256UL
	};
	SIZE_T m_StreamBufferLength;
	SIZE_T m_StreamBufferUsed;
	SIZE_T m_StreamBufferPos;
	BYTE *m_pStreamBuffer;
	SIZE_T m_StreamSendBufferLength;
	HANDLE m_hStreamThread;
	volatile bool m_fStreamEnd;
	CCriticalLock m_StreamLock;
	HANDLE m_hProcessTSEvent;
	DWORD m_DropCount;
	bool LoadSettings();
	bool CreateDummyWindow();
	static CTVBridge *GetThis(HWND hwnd);
	static LRESULT CALLBACK DummyWndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam);
	static LRESULT CALLBACK EventCallback(UINT Event,LPARAM lParam1,LPARAM lParam2,void *pClientData);
	static BOOL CALLBACK StreamCallback(BYTE *pData,void *pClientData);
	bool GetStream(BYTE *pBuffer,SIZE_T *pBufferLength);
	void PurgeStream();
	static DWORD WINAPI StreamThread(LPVOID lpParameter);
public:
	CTVBridge();
	~CTVBridge();
	// CTVTestPlugin
	virtual bool GetPluginInfo(TVTest::PluginInfo *pInfo);
	virtual bool Initialize();
	virtual bool Finalize();
};


CTVBridge::CTVBridge()
	: m_pDTVPlugin(NULL)
	, m_pDTVApp(NULL)
	, m_FriioID(FRIIO_TERRESTRIAL)
	, m_TerrestrialIndex(0)
	, m_BSIndex(1)
	, m_CSIndex(2)
	, m_hwndDummy(NULL)
	, m_StreamBufferLength(DEFAULT_BUFFER_LENGTH)
	, m_pStreamBuffer(NULL)
	, m_StreamSendBufferLength(DEFAULT_SEND_BUFFER_LENGTH)
	, m_hStreamThread(NULL)
	, m_fStreamEnd(false)
	, m_hProcessTSEvent(NULL)
	, m_DropCount(0)
{
	::GetModuleFileName(g_hinstDLL,m_szFriioPluginName,MAX_PATH);
	::PathRenameExtension(m_szFriioPluginName,TEXT(".dll"));
	m_hLib=::LoadLibrary(m_szFriioPluginName);
	if (m_hLib==NULL)
		return;

#if SDK_VERSION>=2
	GetIfacePtr2 pGetIface2=reinterpret_cast<GetIfacePtr2>(::GetProcAddress(m_hLib,"GetIface2"));

	if (pGetIface2==NULL
		|| !pGetIface2(
#if SDK_VERSION>=3
			0x00010500,
#else
			0x00010400,
#endif
			reinterpret_cast<void**>(&m_pDTVPlugin))) {
		::FreeLibrary(m_hLib);
		m_hLib=NULL;
		return;
	}
#else
	GetIfacePtr pGetIface=reinterpret_cast<GetIfacePtr>(::GetProcAddress(m_hLib,"GetIface"));

	if (pGetIface==NULL
		|| (m_pDTVPlugin=reinterpret_cast<IDTVPlugin_*>(pGetIface()))==NULL) {
		::FreeLibrary(m_hLib);
		m_hLib=NULL;
		return;
	}
#endif
	LoadSettings();
}


bool CTVBridge::LoadSettings()
{
	TCHAR szIniFileName[MAX_PATH];

	::GetModuleFileName(g_hinstDLL,szIniFileName,MAX_PATH);
	::lstrcat(szIniFileName,TEXT(".ini"));
	m_FriioID=::GetPrivateProfileInt(TEXT("Settings"),TEXT("FriioID"),FRIIO_TERRESTRIAL,szIniFileName);
	m_TerrestrialIndex=::GetPrivateProfileInt(TEXT("Settings"),TEXT("TerraIndex"),0,szIniFileName);
	m_BSIndex=::GetPrivateProfileInt(TEXT("Settings"),TEXT("BSIndex"),1,szIniFileName);
	m_CSIndex=::GetPrivateProfileInt(TEXT("Settings"),TEXT("CSIndex"),1,szIniFileName);
	m_StreamBufferLength=::GetPrivateProfileInt(TEXT("Settings"),TEXT("BufferLength"),DEFAULT_BUFFER_LENGTH,szIniFileName);
	if (m_StreamBufferLength==0)
		m_StreamBufferLength=DEFAULT_BUFFER_LENGTH;
	m_StreamSendBufferLength=::GetPrivateProfileInt(TEXT("Settings"),TEXT("ProcessBufferLength"),DEFAULT_SEND_BUFFER_LENGTH,szIniFileName);
	if (m_StreamSendBufferLength==0)
		m_StreamSendBufferLength=DEFAULT_SEND_BUFFER_LENGTH;
	return true;
}


CTVBridge::~CTVBridge()
{
	if (m_hLib)
		::FreeLibrary(m_hLib);
	delete m_pDTVApp;
}


bool CTVBridge::GetPluginInfo(TVTest::PluginInfo *pInfo)
{
	if (m_hLib==NULL)
		return false;
	pInfo->Type           = TVTest::PLUGIN_TYPE_NORMAL;
	pInfo->Flags          = TVTest::PLUGIN_FLAG_ENABLEDEFAULT;
	pInfo->pszPluginName  = ::PathFindFileName(m_szFriioPluginName);
	pInfo->pszCopyright   = L"Public Domain";
	pInfo->pszDescription = L"TVTestFriioViewpvOCg";
	return true;
}


bool CTVBridge::Initialize()
{
	if (!CreateDummyWindow())
		return false;
	m_pDTVApp=new CDTVApp(m_pApp,m_FriioID);
	m_pDTVApp->SetSpaceIndex(m_TerrestrialIndex,m_BSIndex,m_CSIndex);
	if (!m_pDTVPlugin->SetHostAPI(reinterpret_cast<IDTVApp*>(m_pDTVApp),
#if SDK_VERSION>=3
								  0x00010500
#elif SDK_VERSION>=2
								  0x00010400
#else
								  0x00010300
#endif
								  )) {
		Finalize();
		return false;
	}
	if (!m_pDTVPlugin->Init(m_hwndDummy)) {
		Finalize();
		return false;
	}
	m_pApp->SetEventCallback(EventCallback,this);
	m_pStreamBuffer=new BYTE[m_StreamBufferLength*188];
	m_StreamBufferUsed=0;
	m_StreamBufferPos=0;
	m_hProcessTSEvent=::CreateEvent(NULL,FALSE,FALSE,NULL);
	m_hStreamThread=::CreateThread(NULL,0,StreamThread,this,0,NULL);
	m_pApp->SetStreamCallback(0,StreamCallback,this);
	return true;
}


bool CTVBridge::Finalize()
{
	if (m_hStreamThread) {
		m_fStreamEnd=true;
		if (::WaitForSingleObject(m_hStreamThread,2000)==WAIT_TIMEOUT)
			::TerminateThread(m_hStreamThread,1);
		::CloseHandle(m_hStreamThread);
		m_hStreamThread=NULL;
	}
	if (m_hProcessTSEvent) {
		::CloseHandle(m_hProcessTSEvent);
		m_hProcessTSEvent=NULL;
	}
	if (m_pStreamBuffer) {
		delete [] m_pStreamBuffer;
		m_pStreamBuffer=NULL;
	}
	if (m_hwndDummy) {
		::SendMessage(m_hwndDummy,WM_CLOSE,0,0);
		::UnregisterClass(m_szClassName,g_hinstDLL);
	}
	if (m_hLib) {
		::FreeLibrary(m_hLib);
		m_hLib=NULL;
	}
	if (m_pDTVApp) {
		delete m_pDTVApp;
		m_pDTVApp=NULL;
	}
	return true;
}


bool CTVBridge::CreateDummyWindow()
{
	if (m_hwndDummy)
		return true;

	WNDCLASS wc;

	::wsprintf(m_szClassName,DUMMY_WINDOW_CLASS,::PathFindFileName(m_szFriioPluginName));
	wc.style=0;
	wc.lpfnWndProc=DummyWndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=g_hinstDLL;
	wc.hIcon=NULL;
	wc.hCursor=NULL;
	wc.hbrBackground=NULL;
	wc.lpszMenuName=NULL;
	wc.lpszClassName=m_szClassName;
	if (::RegisterClass(&wc)==0)
		return false;

	RECT rc;

	::GetWindowRect(m_pApp->GetAppWindow(),&rc);
	if (::CreateWindowEx(WS_EX_LAYERED | WS_EX_TRANSPARENT,m_szClassName,
						 TEXT("Friio HDTV Player"),WS_POPUP,
						 rc.left,rc.top,rc.right-rc.left,rc.bottom-rc.top,
						 m_pApp->GetAppWindow(),NULL,g_hinstDLL,this)==NULL)
		return false;
	::SetLayeredWindowAttributes(m_hwndDummy,0,0,LWA_ALPHA);
	return true;
}


CTVBridge *CTVBridge::GetThis(HWND hwnd)
{
	return reinterpret_cast<CTVBridge*>(::GetWindowLongPtr(hwnd,GWLP_USERDATA));
}


/*
	FriioView̐UăbZ[Wɓ
*/
LRESULT CALLBACK CTVBridge::DummyWndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	switch (uMsg) {
	case WM_CREATE:
		{
			LPCREATESTRUCT pcs=reinterpret_cast<LPCREATESTRUCT>(lParam);
			CTVBridge *pThis=static_cast<CTVBridge*>(pcs->lpCreateParams);

			::SetWindowLongPtr(hwnd,GWLP_USERDATA,reinterpret_cast<LONG_PTR>(pThis));
			pThis->m_hwndDummy=hwnd;
		}
		return 0;

	case WM_STARTREC:
		{
			CTVBridge *pThis=GetThis(hwnd);
			TVTest::RecordInfo RecordInfo;

			TRACE(TEXT("TVBridge : WM_STARTREC %d min / Delay %d sec\n"),LOWORD(wParam),HIWORD(wParam));
			pThis->m_RecordTime=LOWORD(wParam)*60;
			RecordInfo.Mask=TVTest::RECORD_MASK_STARTTIME |
							TVTest::RECORD_MASK_STOPTIME;
			RecordInfo.Flags=0;
			RecordInfo.StartTimeSpec=TVTest::RECORD_START_DELAY;
			RecordInfo.StartTime.Delay=HIWORD(wParam)*1000;
			RecordInfo.StopTimeSpec=TVTest::RECORD_STOP_DURATION;
			RecordInfo.StopTime.Duration=pThis->m_RecordTime*1000;
			return pThis->m_pApp->StartRecord(&RecordInfo);
		}

	case WM_STOPREC:
		TRACE(TEXT("TVBridge : WM_STOPREC\n"));
		return GetThis(hwnd)->m_pApp->StopRecord();

	case WM_CHANGECHAN:
		TRACE(TEXT("TVBridge : WM_CHANGECHAN %ld(0x%lx)\n"),wParam,wParam);
		return GetThis(hwnd)->m_pDTVApp->put_Channel(wParam);

	case WM_ISRECEIVING:
		TRACE(TEXT("TVBridge : WM_ISRECEIVING\n"));
		{
			TVTest::StatusInfo Status;

			if (GetThis(hwnd)->m_pApp->GetStatus(&Status)
					&& Status.SignalLevel>0.0 && Status.BitRate>0)
				return TRUE;
		}
		return FALSE;

	case WM_UPDATECHAN:
		TRACE(TEXT("TVBridge : WM_UPDATECHAN\n"));
		return TRUE;

	case WM_SCREENSHOT:
		TRACE(TEXT("TVBridge : WM_SCREENSHOT\n"));
		return FALSE;

	case WM_SETAUDIO:
		TRACE(TEXT("TVBridge : WM_SETAUDIO %ld %ld\n"),wParam,lParam);
		return GetThis(hwnd)->m_pApp->SetStereoMode(wParam);

	case WM_GETAUDIO:
		TRACE(TEXT("TVBridge : WM_GETAUDIO\n"));
		return MAKELRESULT(GetThis(hwnd)->m_pApp->GetStereoMode(),0);

	case WM_SETASPECT:
		TRACE(TEXT("TVBridge : WM_SETASPECT %ld\n"),lParam);
		{
			CTVBridge *pThis=GetThis(hwnd);
			TVTest::PanScanInfo PanScan;

			if (lParam!=1) {
				TVTest::VideoInfo VideoInfo;

				if (!pThis->m_pApp->GetVideoInfo(&VideoInfo))
					return FALSE;
				PanScan.Type=TVTest::PANSCAN_NONE;
				PanScan.XAspect=VideoInfo.XAspect;
				PanScan.YAspect=VideoInfo.YAspect;
			} else {
				PanScan.Type=TVTest::PANSCAN_NONE;
				PanScan.XAspect=0;
				PanScan.YAspect=0;
			}
			return pThis->m_pApp->SetPanScan(&PanScan);
		}

	case WM_GETASPECT:
		TRACE(TEXT("TVBridge : WM_GETASPECT\n"));
		{
			CTVBridge *pThis=GetThis(hwnd);
			TVTest::PanScanInfo Info;
			TVTest::VideoInfo VideoInfo;

			if (pThis->m_pApp->GetPanScan(&Info)
					&& Info.XAspect>0 && Info.YAspect>0)
				return MAKELRESULT(Info.XAspect,Info.YAspect);
			if (pThis->m_pApp->GetVideoInfo(&VideoInfo))
				return MAKELRESULT(VideoInfo.XAspect,VideoInfo.YAspect);
		}
		return 0;

	case WM_MODIFYREC:
		TRACE(TEXT("TVBridge : WM_MODIFYREC %ld\n"),lParam);
		{
			CTVBridge *pThis=GetThis(hwnd);
			LONG Delta=lParam;
			TVTest::RecordInfo RecordInfo;

			if (Delta>=0 || pThis->m_RecordTime>(DWORD)-Delta)
				pThis->m_RecordTime+=Delta;
			else
				pThis->m_RecordTime=0;
			RecordInfo.Mask=TVTest::RECORD_MASK_STOPTIME;
			RecordInfo.Flags=0;
			RecordInfo.StopTimeSpec=TVTest::RECORD_STOP_DURATION;
			RecordInfo.StopTime.Duration=pThis->m_RecordTime*1000;
			return pThis->m_pApp->ModifyRecord(&RecordInfo);
		}

	case WM_GETRECLEN:
		TRACE(TEXT("TVBridge : WM_GETRECLEN\n"));
		{
			TVTest::RecordStatusInfo RecordStatus;

			if (GetThis(hwnd)->m_pApp->GetRecordStatus(&RecordStatus)
					&& RecordStatus.Status!=TVTest::RECORD_STATUS_NOTRECORDING)
				return (RecordStatus.RecordTime+RecordStatus.PauseTime)/1000;
		}
		return 0;

	case WM_GETRECLEFT:
		TRACE(TEXT("TVBridge : WM_GETRECLEFT\n"));
		{
			TVTest::RecordStatusInfo RecordStatus;
			DWORD RemainTime=0;

			if (GetThis(hwnd)->m_pApp->GetRecordStatus(&RecordStatus)
				&& RecordStatus.Status!=TVTest::RECORD_STATUS_NOTRECORDING) {
				DWORD RecordTime=RecordStatus.RecordTime+RecordStatus.PauseTime;

				if (RecordStatus.StopTimeSpec==TVTest::RECORD_STOP_TIME) {
					SYSTEMTIME st;
					FILETIME ft;
					ULARGE_INTEGER Cur,Stop;

					::GetLocalTime(&st);
					::SystemTimeToFileTime(&st,&ft);
					Cur.LowPart=ft.dwLowDateTime;
					Cur.HighPart=ft.dwHighDateTime;
					Stop.LowPart=RecordStatus.StopTime.Time.dwLowDateTime;
					Stop.HighPart=RecordStatus.StopTime.Time.dwHighDateTime;
					if (Cur.QuadPart<Stop.QuadPart)
						RemainTime=(DWORD)((Stop.QuadPart-Cur.QuadPart)/10000000);
				} else if (RecordStatus.StopTimeSpec==TVTest::RECORD_STOP_DURATION) {
					if (RecordTime<=RecordStatus.StopTime.Duration)
						RemainTime=
							(DWORD)((RecordStatus.StopTime.Duration-RecordTime)/1000);
				} else {
					RemainTime=(0xFFFFFFFFUL-RecordTime)/1000;
				}
			}
			return RemainTime;
		}

	case WM_GETFRIIOID:
		TRACE(TEXT("TVBridge : WM_GETFRIIOID\n"));
		return GetThis(hwnd)->m_FriioID;

	case WM_LNBPOWER:
		TRACE(TEXT("TVBridge : WM_LNBPOWER\n"));
		return FALSE;

	case WM_PROGRAMS_UPDATED:
		TRACE(TEXT("TVBridge : WM_PROGRAMS_UPDATE\n"));
		return TRUE;

	case WM_DESTROY:
		{
			CTVBridge *pThis=GetThis(hwnd);

			pThis->m_hwndDummy=NULL;
		}
		return 0;
	}
	return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
}


struct ShowWindowInfo {
	HWND hwnd;
	bool fShow;
};

BOOL CALLBACK ShowWindowProc(HWND hwnd,LPARAM lParam)
{
	ShowWindowInfo *pInfo=reinterpret_cast<ShowWindowInfo*>(lParam);

	if (::GetWindow(hwnd,GW_OWNER)==pInfo->hwnd)
		::ShowWindow(hwnd,pInfo->fShow?SW_SHOW:SW_HIDE);
	return TRUE;
}


LRESULT CALLBACK CTVBridge::EventCallback(UINT Event,LPARAM lParam1,LPARAM lParam2,void *pClientData)
{
	CTVBridge *pThis=static_cast<CTVBridge*>(pClientData);

	switch (Event) {
	case TVTest::EVENT_PLUGINENABLE:
		{
			bool fEnable=lParam1!=0;
			ShowWindowInfo Info;

			Info.hwnd=pThis->m_hwndDummy;
			Info.fShow=fEnable;
			::EnumWindows(ShowWindowProc,reinterpret_cast<LPARAM>(&Info));
		}
		return TRUE;

	case TVTest::EVENT_CHANNELCHANGE:
	case TVTest::EVENT_DRIVERCHANGE:
		pThis->PurgeStream();
		return TRUE;

	case TVTest::EVENT_SERVICECHANGE:
	case TVTest::EVENT_SERVICEUPDATE:
		pThis->m_pDTVPlugin->OnStreamChange(pThis->m_pDTVApp->get_Channel(),
											pThis->m_pDTVApp->get_ServiceID());
		return TRUE;
	}
	return 0;
}


BOOL CALLBACK CTVBridge::StreamCallback(BYTE *pData,void *pClientData)
{
	CTVBridge *pThis=static_cast<CTVBridge*>(pClientData);

	pThis->m_StreamLock.Lock();
	if (pThis->m_StreamBufferUsed==pThis->m_StreamBufferLength) {
		pThis->m_DropCount++;
		pThis->m_StreamLock.Unlock();
		return TRUE;
	}
	::CopyMemory(pThis->m_pStreamBuffer+(pThis->m_StreamBufferPos+pThis->m_StreamBufferUsed)%pThis->m_StreamBufferLength*188,pData,188);
	pThis->m_StreamBufferUsed++;
	pThis->m_StreamLock.Unlock();
	return TRUE;
}


bool CTVBridge::GetStream(BYTE *pBuffer,SIZE_T *pBufferLength)
{
	m_StreamLock.Lock();
	if (m_StreamBufferUsed==0) {
		m_StreamLock.Unlock();
		return false;
	}
	SIZE_T Size=min(*pBufferLength,m_StreamBufferUsed);
	if (Size>m_StreamBufferLength-m_StreamBufferPos)
		Size=m_StreamBufferLength-m_StreamBufferPos;
	::CopyMemory(pBuffer,m_pStreamBuffer+m_StreamBufferPos*188,Size*188);
	*pBufferLength=Size;
	m_StreamBufferPos+=Size;
	if (m_StreamBufferPos==m_StreamBufferLength)
		m_StreamBufferPos=0;
	m_StreamBufferUsed-=Size;
	m_StreamLock.Unlock();
	return true;
}


void CTVBridge::PurgeStream()
{
	m_StreamLock.Lock();
	m_StreamBufferUsed=0;
	m_StreamBufferPos=0;
	m_StreamLock.Unlock();
	::ResetEvent(m_hProcessTSEvent);
	::WaitForSingleObject(m_hProcessTSEvent,INFINITE);
}


DWORD WINAPI CTVBridge::StreamThread(LPVOID lpParameter)
{
	CTVBridge *pThis=static_cast<CTVBridge*>(lpParameter);
	BYTE *pBuffer=new BYTE[pThis->m_StreamSendBufferLength*188];

	while (!pThis->m_fStreamEnd) {
		SIZE_T Length=pThis->m_StreamSendBufferLength;

		if (pThis->GetStream(pBuffer,&Length)) {
			pThis->m_pDTVPlugin->ProcessTS(pBuffer,Length*188);
			::SetEvent(pThis->m_hProcessTSEvent);
		} else {
			::SetEvent(pThis->m_hProcessTSEvent);
			::Sleep(5);
		}
	}
	delete [] pBuffer;
	return 0;
}




TVTest::CTVTestPlugin *CreatePluginClass()
{
	return new CTVBridge;
}
